#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import os
import tempfile
from sos_notebook.test_utils import NotebookTest


class TestInterface(NotebookTest):

    def test_prompt_color(self, notebook):
        '''test color of input and output prompt'''
        idx = notebook.call(
            '''\
            cat('this is R')
            ''', kernel="R")
        assert [220, 220, 218] == notebook.get_input_backgroundColor(idx)
        assert [220, 220, 218] == notebook.get_output_backgroundColor(idx)

    def test_cd(self, notebook):
        '''Support for change of directory with magic %cd'''
        output1 = notebook.check_output('getwd()', kernel="R")
        notebook.call('%cd ..', kernel="SoS")
        output2 = notebook.check_output('getwd()', kernel="R")
        assert len(output1) > len(output2)
        assert output1.strip("'").startswith(output2.strip("'"))
        #
        # cd to a specific directory
        tmpdir = os.path.join(tempfile.gettempdir(), 'somedir')
        os.makedirs(tmpdir, exist_ok=True)
        notebook.call(f'%cd "{tmpdir}"', kernel="SoS")
        output = notebook.check_output('cat(getwd())', kernel="R")
        assert os.path.realpath(tmpdir) == os.path.realpath(output)

    def test_var_names(self, notebook):
        '''Test get/put variables with strange names'''
        # _a_1 => .a_1 in R
        notebook.call(
            '''\
            %put _a_1 --to R
            _a_1 = 5
            ''',
            kernel="SoS",
            expect_error=True)
        assert '5' == notebook.check_output('.a_1', kernel='R')
        # .a.1 => _a_1 in Python (SoS)
        notebook.call(
            '''\
            %put .a.1
            .a.1 = 500
            ''',
            kernel="R",
            expect_error=True)
        assert '500' == notebook.check_output('_a_1', kernel='SoS')

    def test_expand(self, notebook):
        '''Test %expand --in R'''
        notebook.call('var = 100', kernel="R")
        assert 'value is 103' in notebook.check_output(
            '''\
            %expand --in R
            value is {var + 3}
            ''',
            kernel='Markdown')
        assert 'value is 102' in notebook.check_output(
            '''\
            %expand `r ` --in R
            value is `r var + 2`
            ''',
            kernel='Markdown')

    def test_preview(self, notebook):
        '''Test support for %preview'''
        output = notebook.check_output(
            '''\
            %preview -n var
            var = seq(1, 1000)
            ''',
            kernel="R")
        # in a normal var output, 100 would be printed. The preview version would show
        # type and some of the items in the format of
        #   int [1:1000] 1 2 3 4 5 6 7 8 9 10 ...
        assert 'int' in output and '3' in output and '9' in output and '111' not in output
        #
        # return 'Unknown variable' for unknown variable
        assert 'Unknown variable' in notebook.check_output(
            '%preview -n unknown_var', kernel="R")
        #
        # return 'Unknown variable for expression
        assert 'Unknown variable' in notebook.check_output(
            '%preview -n var[1]', kernel="R")

    def test_sessioninfo(self, notebook):
        '''test support for %sessioninfo'''
        notebook.call("cat('this is R')", kernel="R")
        assert 'R version' in notebook.check_output(
            '%sessioninfo', kernel="SoS")



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

import numpy
import pandas
import re
import tempfile

from collections import Sequence
from sos.utils import short_repr, env
from IPython.core.error import UsageError

from ._version import __version__


def homogeneous_type(seq):
    iseq = iter(seq)
    first_type = type(next(iseq))
    if first_type in (int, float):
        return True if all(isinstance(x, (int, float)) for x in iseq) else False
    else:
        return True if all(isinstance(x, first_type) for x in iseq) else False


# make the SoS dict key name to be valid in R list
def make_name(name):
    if name.isalpha():
        return name
    # the best way to detect an empty string is `if not {string}`
    if not name or not name[0].isalpha():
        name = 'X' + name
    return re.sub(r'\W', '_', name)


#
#  support for %get
#
#  Converting a Python object to a R expression that will be executed
#  by the R kernel.
#
#

# use dictionary to remove duplicated warnings
_R_repr_warnings = {}


def _R_repr(obj, processed=None):
    global _R_repr_warnings
    if isinstance(obj, bool):
        return 'TRUE' if obj else 'FALSE'
    elif isinstance(obj, (int, str)):
        return repr(obj)
    elif isinstance(obj, float):
        if numpy.isnan(obj):
            return 'NaN'
        elif numpy.isinf(obj):
            return 'Inf'
        else:
            return repr(obj)
    elif isinstance(obj, complex):
        return 'complex(real = ' + str(obj.real) + ', imaginary = ' + str(
            obj.imag) + ')'
    elif isinstance(obj, Sequence):
        if len(obj) == 0:
            return 'c()'
        # if the data is of homogeneous type, let us use c()
        # otherwise use list()
        # this can be confusion but list can be difficult to handle
        if homogeneous_type(obj):
            return 'c(' + ','.join(_R_repr(x) for x in obj) + ')'
        else:
            return 'list(' + ','.join(_R_repr(x) for x in obj) + ')'
    elif obj is None:
        return 'NULL'
    elif isinstance(obj, dict):
        if processed:
            if id(obj) in processed:
                return 'NULL'
        else:
            processed = set()
        processed.add(id(obj))
        return 'list(' + ','.join(
            '{}={}'.format(make_name(str(x)), _R_repr(y, processed))
            for x, y in obj.items()) + ')'
    elif isinstance(obj, set):
        return 'list(' + ','.join(_R_repr(x) for x in obj) + ')'
    elif isinstance(
            obj, (numpy.intc, numpy.intp, numpy.int8, numpy.int16, numpy.int32,
                  numpy.int64, numpy.uint8, numpy.uint16, numpy.uint32,
                  numpy.uint64, numpy.float16, numpy.float32, numpy.float64)):
        return repr(obj)
    elif isinstance(obj, numpy.matrixlib.defmatrix.matrix):
        try:
            import feather
        except ImportError:
            raise UsageError(
                'The feather-format module is required to pass numpy matrix as R matrix'
                'See https://github.com/wesm/feather/tree/master/python for details.'
            )
        feather_tmp_ = tempfile.NamedTemporaryFile(
            suffix='.feather', delete=False).name
        feather.write_dataframe(pandas.DataFrame(obj).copy(), feather_tmp_)
        return 'data.matrix(..read.feather({!r}))'.format(feather_tmp_)
    elif isinstance(obj, numpy.ndarray):
        if obj.ndim == 1:
            return 'array(c(' + ','.join(_R_repr(x) for x in obj) + '))'
        else:
            return 'array(' + 'c(' + ','.join(
                repr(x)
                for x in obj.swapaxes(obj.ndim - 2, obj.ndim - 1).flatten(
                    order='C')) + ')' + ', dim=(' + 'rev(c' + repr(
                        obj.swapaxes(obj.ndim - 2, obj.ndim - 1).shape) + ')))'
    elif isinstance(obj, pandas.DataFrame):
        try:
            import feather
        except ImportError:
            raise UsageError(
                'The feather-format module is required to pass pandas DataFrame as R data.frame'
                'See https://github.com/wesm/feather/tree/master/python for details.'
            )
        feather_tmp_ = tempfile.NamedTemporaryFile(
            suffix='.feather', delete=False).name
        try:
            data = obj.copy()
            # if the dataframe has index, it would not be transferred due to limitations
            # of feather. We will have to do something to save the index separately and
            # recreate it. (#397)
            if isinstance(data.index, pandas.Index):
                df_index = list(data.index)
                if len(df_index) != len(set(df_index)):
                    df_index = None
                    _R_repr_warnings[
                        'Index is ignored because R dataframe does not accept non-unique row names.'] = 1
            elif not isinstance(data.index, pandas.RangeIndex):
                # we should give a warning here
                df_index = None
            feather.write_dataframe(data, feather_tmp_)
        except Exception:
            # if data cannot be written, we try to manipulate data
            # frame to have consistent types and try again
            for c in data.columns:
                if not homogeneous_type(data[c]):
                    data[c] = [str(x) for x in data[c]]
            feather.write_dataframe(data, feather_tmp_)
            # use {!r} for path because the string might contain c:\ which needs to be
            # double quoted.
        return '..read.feather({!r}, index={})'.format(feather_tmp_,
                                                       _R_repr(df_index))
    elif isinstance(obj, pandas.Series):
        dat = list(obj.values)
        ind = list(obj.index.values)
        return 'setNames(' + 'c(' + ','.join(
            _R_repr(x) for x in dat) + ')' + ',c(' + ','.join(
                _R_repr(y) for y in ind) + '))'
    else:
        _R_repr_warnings[
            'Unsupported datatype {}. Variable is set to NULL'.format(
                short_repr(obj))] = 1
        return 'NULL'


# R    length (n)    Python
# NULL        None
# logical    1    boolean
# integer    1    integer
# numeric    1    double
# character    1    unicode
# logical    n > 1    array
# integer    n > 1    array
# numeric    n > 1    list
# character    n > 1    list
# list without names    n > 0    list
# list with names    n > 0    dict
# matrix    n > 0    array
# data.frame    n > 0    DataFrame

R_init_statements = r'''
..py.repr.logical.1 <- function(obj) {
    if(obj)
        'True'
    else
        'False'
}
..py.repr.integer.1 <- function(obj) {
    as.character(obj)
}
..py.repr.double.1 <- function(obj) {
    if (is.nan(obj)) {
      'numpy.nan'
    } else if (is.infinite(obj)) {
      'float("inf")'
    } else {
        as.character(obj)
    }
}
..py.repr.complex.1 <- function(obj) {
    rl = Re(obj)
    im = Im(obj)
    paste0('complex(', rl, ',', im, ')')
}
..py.repr.character.1 <- function(obj) {
    paste0('r"""', obj, '"""')
}
..has.row.names <- function(df) {
  !all(row.names(df)==seq(1, nrow(df)))
}
..py.repr.dataframe <- function(obj) {
    if (!require("arrow")) {
        install.packages('arrow', repos='https://cran.r-project.org')
        }
    library(arrow)
    tf = tempfile('arrow')
    write_feather(obj, tf)
    if (..has.row.names(obj)) {
        paste0("read_dataframe(r'", tf, "').set_index(pandas.Index(", ..py.repr(row.names(obj)),"))")
    } else {
        paste0("read_dataframe(r'", tf, "')")
    }
}
..py.repr.matrix <- function(obj) {
    if (!require("arrow")) {
        install.packages('arrow', repos='https://cran.r-project.org')
        }
    library(arrow)
    tf = tempfile('arrow')
    write_feather(as.data.frame(obj), tf)
    if (..has.row.names(obj)) {
        paste0("read_dataframe(r'", tf, "').set_index(pandas.Index(", ..py.repr(row.names(obj)),")).values")
    } else {
        paste0("read_dataframe(r'", tf, "').values")
    }
}
..py.repr.array.numer <- function(obj) {
    paste0("numpy.array(", "[", paste(obj, collapse = ","), "]).", paste0("reshape([",
                                                                        paste0(rev(dim(obj)), collapse = ","), "]).", paste0("swapaxes(",
                                                                                                                             length(dim(obj)) - 2, ",", length(dim(obj)) - 1, ")")))
}
..py.repr.array.char <- function(obj) {
    paste0("numpy.array(", "[", paste0( paste0("eval('", ..py.repr.character.1(obj), "')", collapse=',')), "]).", paste0("reshape([",
                                                                                                          paste0(rev(dim(obj)), collapse = ","), "]).", paste0("swapaxes(",
                                                                                                                                                               length(dim(obj)) - 2, ",", length(dim(obj)) - 1, ")")))
}
..py.repr.array.logical <- function(obj) {
  paste0("numpy.array(", "[", paste0( paste0("eval('", apply(obj,c(1:length(dim(obj))),..py.repr.logical.1), "')", collapse=',')), "]).", paste0("reshape([",
                                                                                                                     paste0(rev(dim(obj)), collapse = ","), "]).", paste0("swapaxes(",
                                                                                                                                                                          length(dim(obj)) - 2, ",", length(dim(obj)) - 1, ")")))
}
..py.repr.n <- function(obj) {
    paste("[",
        paste(sapply(obj, ..py.repr), collapse=','),
        "]")
}
..py.repr <- function(obj) {
    if (is.matrix(obj)) {
      ..py.repr.matrix(obj)
    } else if (is.data.frame(obj)) {
      ..py.repr.dataframe(obj)
    } else if (is.list(obj)) {
      # if the list has no name
      if (is.null(names(obj)))
        ..py.repr.n(obj)
      else {
        paste("dict([",
              paste(sapply(names(obj), function (x)
                paste0("(", shQuote(gsub("\\.", "_", as.character(x))), ",", ..py.repr(obj[[x]]), ")" )),
                collapse=','),
              "])")
        }
    } else if (is.array(obj)) {
      if (is.character(obj))
        ..py.repr.array.char(obj)
      else if (is.logical(obj))
        ..py.repr.array.logical(obj)
      else
        ..py.repr.array.numer(obj)
    } else if (is.null(obj)) {
      'None'
    } else if (is.integer(obj)) {
        # if the vector has no name
        if (is.null(names(obj)))
          if (length(obj) == 1)
            ..py.repr.integer.1(obj)
          else
            paste("[", paste(obj, collapse=','), "]")
        else
          paste0("pandas.Series(", "[", paste(unname(obj), collapse=','), "],", paste0("[", paste0(sapply(names(obj), ..py.repr.character.1), collapse=','), "]"), ")")
    } else if (is.complex(obj)) {
        # if the vector has no name
        if (is.null(names(obj)))
          if (length(obj) == 1)
            ..py.repr.complex.1(obj)
          else
            paste("[", paste(sapply(obj, ..py.repr.complex.1), collapse=','), "]")
        else
          paste0("pandas.Series(", "[", paste(sapply(unname(obj), ..py.repr.complex.1), collapse=','), "],", paste0("[", paste0(sapply(names(obj), ..py.repr.character.1), collapse=','), "]"), ")")
    } else if (is.double(obj)){
        if (is.null(names(obj))) {
          if (length(obj) == 1) {
            ..py.repr.double.1(obj)
          } else {
            paste("[", paste(sapply(obj, ..py.repr.double.1), collapse=','), "]")
          }
        } else {
          paste0("pandas.Series(", "[", paste(unname(obj), collapse=','), "],", paste0("[", paste0(sapply(names(obj), ..py.repr.character.1), collapse=','), "]"), ")")
        }
    } else if (is.character(obj)) {
        # if the vector has no name
        if (is.null(names(obj)))
          if (length(obj) == 1)
            ..py.repr.character.1(obj)
          else
            paste("[", paste(sapply(obj, ..py.repr.character.1), collapse=','), "]")
        else
          paste0("pandas.Series(", "[", paste(sapply(unname(obj), ..py.repr.character.1), collapse=','), "],", paste0("[", paste0(sapply(names(obj), ..py.repr.character.1), collapse=','), "]"), ")")
    } else if (is.logical(obj)) {
      # if the vector has no name
        if (is.na(obj)) {
            'numpy.nan'
        } else if (is.null(names(obj)))
          if (length(obj) == 1)
            ..py.repr.logical.1(obj)
          else
            ..py.repr.n(obj)
        else
          paste0("pandas.Series(", "[", paste(sapply(unname(obj), ..py.repr.logical.1), collapse=','), "],", paste0("[", paste0(sapply(names(obj), ..py.repr.character.1), collapse=','), "]"), ")")
    } else {
      "'Untransferrable variable'"
    }
}
..read.feather <- function(filename, index=NULL) {
    if (! suppressMessages(suppressWarnings(require("arrow", quietly = TRUE)))) {
      try(install.packages('arrow', repos='https://cran.r-project.org'), silent=TRUE)
      if (!suppressMessages(suppressWarnings(require("arrow"))))
        stop('Failed to install arrow library')
    }
    suppressPackageStartupMessages(library(arrow, quietly = TRUE))
    data = as.data.frame(read_feather(filename))
    if (!is.null(index))
      rownames(data) <- index
    for (cn in names(data)[sapply(data,is, class2="integer64")]) {
      data[[cn]] <- tryCatch( {
          as.integer(data[[cn]])
      }, warning = function (w) {
          message("out of range integer column is converted to numeric")
          as.numeric(data[[cn]])
      }
    )
    }
    return(data)
}
..sos.preview <- function(name) {
    tryCatch( str(get(name)), error = function(err) { cat(paste('Unknown variable', name)) })
}
..sos.expand <- function(text, sigil) {
    if (! suppressMessages(suppressWarnings(require("knitr", quietly = TRUE)))) {
      try(install.packages('knitr', repos='https://cran.r-project.org'), silent=TRUE)
      if (!suppressMessages(suppressWarnings(require("knitr"))))
        stop('Failed to install knitr library')
    }
    suppressPackageStartupMessages(library(knitr, quietly = TRUE))
    cat(knit_expand(text=text, delim=sigil))
}
'''


class sos_R:
    background_color = '#DCDCDA'
    supported_kernels = {'R': ['ir']}
    options = {'assignment_pattern': r'^\s*([_A-Za-z0-9\.]+)\s*(=|<-).*$'}
    cd_command = 'setwd({dir!r})'
    __version__ = __version__

    def __init__(self, sos_kernel, kernel_name='ir'):
        self.sos_kernel = sos_kernel
        self.kernel_name = kernel_name
        self.init_statements = R_init_statements

    def get_vars(self, names):
        global _R_repr_warnings
        for name in names:
            if name.startswith('_'):
                self.sos_kernel.warn(
                    f'Variable {name} is passed from SoS to kernel {self.kernel_name} as {"." + name[1:]}'
                )
                newname = '.' + name[1:]
            else:
                newname = name
            _R_repr_warnings = {}
            r_repr = _R_repr(env.sos_dict[name])
            if _R_repr_warnings:
                self.sos_kernel.warn('\n'.join(_R_repr_warnings.keys()))
                _R_repr_warnings = {}
            env.log_to_file('VARIABLE', r_repr)
            self.sos_kernel.run_cell(
                f'{newname} <- {r_repr}',
                True,
                False,
                on_error=f'Failed to get variable {name} to R')

    def put_vars(self, items, to_kernel=None):
        # first let us get all variables with names starting with sos
        response = self.sos_kernel.get_response(
            'cat(..py.repr(ls()))', ('stream',), name=('stdout',))[0][1]
        all_vars = eval(response['text'])
        all_vars = [all_vars] if isinstance(all_vars, str) else all_vars

        items += [x for x in all_vars if x.startswith('sos')]

        for item in items:
            if '.' in item:
                self.sos_kernel.warn(
                    f'Variable {item} is put to SoS as {item.replace(".", "_")}'
                )

        if not items:
            return {}

        py_repr = f'cat(..py.repr(list({",".join("{0}={0}".format(x) for x in items)})))'
        response = self.sos_kernel.get_response(
            py_repr, ('stream',), name=('stdout',))[0][1]
        expr = response['text']

        if to_kernel in ('Python2', 'Python3'):
            # directly to python3
            return '{}\n{}\n{}\nglobals().update({})'.format(
                'from feather import read_dataframe\n'
                if 'read_dataframe' in expr else '',
                'import numpy' if 'numpy' in expr else '',
                'import pandas' if 'pandas' in expr else '', expr)
        # to sos or any other kernel
        else:
            # irkernel (since the new version) does not produce execute_result, only
            # display_data
            try:
                if 'read_dataframe' in expr:
                    # imported to be used by eval
                    from feather import read_dataframe
                    # suppress flakes warning
                    assert read_dataframe
                # evaluate as raw string to correctly handle \\ etc
                return eval(expr)
            except Exception as e:
                self.sos_kernel.warn(f'Failed to evaluate {expr!r}: {e}')
                return None

    def expand(self, text, sigil):
        if '"' in sigil:
            self.sos_kernel.warn(f'Unacceptable delimiter {sigil}')
            return text
        try:
            text = text.replace('"', r'\"')
            l, r = sigil.split(' ')
            # in the case of "`r", we actually use "`r " as left delimiter.
            if l[-1].isalpha():
                l = l + ' '
            if r[0].isalpha():
                r = ' ' + r
            return self.sos_kernel.get_response(
                f'..sos.expand("{text}", c("{l}", "{r}"))', ('stream',),
                name=('stdout',))[0][1]['text']
        except Exception:
            err_msg = self.sos_kernel.get_response(
                f'..sos.expand("{text}", c("{l}", "{r}"))', ('error',),
                name=('evalue',))[0][1]['evalue']
            self.sos_kernel.warn(
                f'Failed to expand {text} with sigil {sigil}: {err_msg}')
            return text

    def preview(self, item):
        # return the preview of variable.
        try:
            return "", self.sos_kernel.get_response(
                f'..sos.preview("{item}")', ('stream',),
                name=('stdout',))[0][1]['text']
        except Exception as e:
            env.log_to_file('VARIABLE', f'Preview of {item} failed: {e}')
            return None

    def sessioninfo(self):
        response = self.sos_kernel.get_response(
            r'cat(paste(capture.output(sessionInfo()), collapse="\n"))',
            ('stream',),
            name=('stdout',))[0]
        return response[1]['text']



#!/usr/bin/env python3
#
# Copyright (c) Bo Peng and the University of Texas MD Anderson Cancer Center
# Distributed under the terms of the 3-clause BSD License.

from sos_notebook.test_utils import NotebookTest
import random


class TestDataExchange(NotebookTest):

    def _var_name(self):
        if not hasattr(self, '_var_idx'):
            self._var_idx = 0
        self._var_idx += 1
        return f'var{self._var_idx}'

    def get_from_SoS(self, notebook, sos_expr):
        var_name = self._var_name()
        notebook.call(f'{var_name} = {sos_expr}', kernel='SoS')
        return notebook.check_output(
            f'''\
            %get {var_name}
            {var_name}''',
            kernel='R')

    def put_to_SoS(self, notebook, r_expr):
        var_name = self._var_name()
        notebook.call(
            f'''\
            %put {var_name}
            {var_name} <- {r_expr}
            ''',
            kernel='R')
        return notebook.check_output(f'print(repr({var_name}))', kernel='SoS')

    def test_get_none(self, notebook):
        assert 'NULL' == self.get_from_SoS(notebook, 'None')

    def test_put_null(self, notebook):
        assert 'None' == self.put_to_SoS(notebook, 'NULL')

    def test_get_numpy_inf(self, notebook):
        notebook.call('import numpy', kernel='SoS')
        assert 'Inf' == self.get_from_SoS(notebook, 'numpy.inf')

    def test_put_inf(self, notebook):
        assert 'inf' == self.put_to_SoS(notebook, 'Inf')

    def test_get_int(self, notebook):
        assert '123' == self.get_from_SoS(notebook, '123')
        assert '1234567891234' == self.get_from_SoS(notebook, '1234567891234')
        # longer int does not gurantee precision any more, I am not sure
        # if the following observation will be consistent (784 vs 789)
        assert '123456789123456784' == self.get_from_SoS(
            notebook, '123456789123456789')

    def test_put_int(self, notebook):
        assert '123' == self.put_to_SoS(notebook, '123')
        assert '1234567891234' == self.put_to_SoS(notebook, '1234567891234')
        # longer int does not gurantee precision any more, I am not sure
        # if the following observation will be consistent (784 vs 789)
        assert '123456789123456784' == self.put_to_SoS(notebook,
                                                       '123456789123456789')

    def test_get_double(self, notebook):
        # FIXME: can we improve the precision here? Passing float as string
        # is certainly a bad idea.
        val = str(random.random())
        assert abs(float(val) - float(self.get_from_SoS(notebook, val))) < 1e-10

    def test_put_double(self, notebook):
        val = str(random.random())
        assert abs(float(val) - float(self.put_to_SoS(notebook, val))) < 1e-10

    def test_get_logic(self, notebook):
        assert 'TRUE' == self.get_from_SoS(notebook, 'True')
        assert 'FALSE' == self.get_from_SoS(notebook, 'False')

    def test_put_logic(self, notebook):
        assert 'True' == self.put_to_SoS(notebook, 'TRUE')
        assert 'False' == self.put_to_SoS(notebook, 'FALSE')

    def test_get_num_array(self, notebook):
        integers = [str(random.randint(1, 100000)) for x in range(10)]
        assert integers[0] == self.get_from_SoS(notebook, f'[{integers[0]}]')
        output = self.get_from_SoS(notebook, f"[{','.join(integers)}]")
        assert all(x in output for x in integers)
        #
        floats = ['1.34', '2.4']        
        assert floats[0] == self.get_from_SoS(notebook, f'[{floats[0]}]')
        output = self.get_from_SoS(notebook, f"[{','.join(floats)}]")
        assert all(x in output for x in floats)

    def test_put_num_array(self, notebook):
        integers = [str(random.randint(1, 100000)) for x in range(10)]
        assert integers[0] == self.put_to_SoS(notebook, f'c({integers[0]})')
        output = self.put_to_SoS(notebook, f'c({",".join(integers)})')
        assert all(x in output for x in integers)
        #
        floats = ['1232.43', '343.444', '90']
        assert floats[0] == self.put_to_SoS(notebook, f'c({floats[0]})')
        output = self.put_to_SoS(notebook, f'c({",".join(floats)})')
        assert all(x in output for x in floats)
        

    def test_get_logic_array(self, notebook):
        assert 'TRUE' == self.get_from_SoS(notebook, '[True]')
        output = self.get_from_SoS(notebook,
                                                      '[True, False, True]')
        assert output.count('TRUE') == 2                                                      
        assert output.count('FALSE') == 1

    def test_put_logic_array(self, notebook):
        # Note that single element numeric array is treated as single value
        assert 'True' == self.put_to_SoS(notebook, 'c(TRUE)')
        assert '[True, False, True]' == self.put_to_SoS(notebook,
                                                        'c(TRUE, FALSE, TRUE)')

    def test_get_str(self, notebook):
        assert "'ab c d'" == self.get_from_SoS(notebook, "'ab c d'")
        assert "'ab\\td'" == self.get_from_SoS(notebook, r"'ab\td'")

    def test_put_str(self, notebook):
        assert "'ab c d'" == self.put_to_SoS(notebook, "'ab c d'")
        assert "'ab\\td'" == self.put_to_SoS(notebook, r"'ab\td'")

    def test_get_mixed_list(self, notebook):
        assert "1.4\nTRUE\n'asd'" == self.get_from_SoS(notebook,
                                                       '[1.4, True, "asd"]')

    def test_put_mixed_list(self, notebook):
        # R does not have mixed list, it just convert everything to string.
        assert "['1.4', 'TRUE', 'asd']" == self.put_to_SoS(
            notebook, 'c(1.4, TRUE, "asd")')

    def test_get_dict(self, notebook):
        # Python does not have named ordered list, so get dictionary
        assert "$a\n1\n$b\n2\n$c\n'3'" == self.get_from_SoS(
            notebook, "dict(a=1, b=2, c='3')")

    def test_put_named_list(self, notebook):
        assert "{'a': 1, 'b': 2, 'c': '3'}" == self.put_to_SoS(
            notebook, "list(a=1, b=2, c='3')")

    def test_get_set(self, notebook):
        output = self.get_from_SoS(notebook, "{1.5, 'abc'}")
        assert "1.5\n'abc'" == output or "'abc'\n1.5" == output

    def test_put_unnamed_list(self, notebook):
        output = self.put_to_SoS(notebook, "list(1.5, 'abc')")
        assert "[1.5, 'abc']" == output or "['abc', 1.5]" == output

    def test_get_complex(self, notebook):
        assert "1+2.2i" == self.get_from_SoS(notebook, "complex(1, 2.2)")

    def test_put_complex(self, notebook):
        assert "(1+2.2j)" == self.put_to_SoS(notebook,
                                             "complex(real=1, imaginary=2.2)")

    def test_get_recursive(self, notebook):
        assert "$a\n1\n$b\n$c\n3\n$d\n'whatever'" == self.get_from_SoS(
            notebook, "{'a': 1, 'b': {'c': 3, 'd': 'whatever'}}")

    def test_put_recursive(self, notebook):
        assert "{'a': 1, 'b': {'c': 3, 'd': 'whatever'}}" == self.put_to_SoS(
            notebook, "list(a=1, b=list(c=3, d='whatever'))")

    def test_get_series(self, notebook):
        notebook.call('import pandas as pd', kernel='SoS')
        nums = [str(random.randint(1, 10000)) for x in range(10)]
        output = self.get_from_SoS(notebook, f'pd.Series([{",".join(nums)}])')
        for num in nums:
            assert num in output
        

    def test_put_series(self, notebook):
        output = self.put_to_SoS(notebook,
                                 "setNames(c(11, 22, 33), c('a', 'b', 'c'))")
        assert 'a    11' in output and 'b    22' in output and 'c    33' in output

    def test_get_matrix(self, notebook):
        notebook.call('import numpy as np', kernel='SoS')
        assert "0 1\n1 2\n3 4" in self.get_from_SoS(notebook,
                                                    'np.matrix([[1,2],[3,4]])')

    def test_put_matrix(self, notebook):
        output = self.put_to_SoS(notebook,
                                 "matrix(c(2, 4, 3, 1, 5, 7), nrow=2)")
        assert 'array' in output and '[2., 3., 5.]' in output and '[4., 1., 7.]' in output

    def test_get_dataframe(self, notebook):
        notebook.call(
            '''\
            %put df --to R
            import pandas as pd
            import numpy as np
            arr = np.random.randn(1000)
            arr[::10] = np.nan
            df = pd.DataFrame({'column_{0}'.format(i): arr for i in range(10)})
            ''',
            kernel='SoS')
        assert '1000' == notebook.check_output('dim(df)[1]', kernel='R')
        assert '10' == notebook.check_output('dim(df)[2]', kernel='R')

    def test_get_dataframe_bigint(self, notebook):
        notebook.call(
            '''\
            %put df --to R
            import pandas as pd
            import numpy as np
            df = pd.DataFrame(dict(a=[2147483648, 1], b=[2, 3]))
            ''',
            kernel='SoS',
            expect_error=False)
        assert '2147483648' in notebook.check_output('df[[1, 1]]', kernel='R')
        assert 'numeric' in notebook.check_output('class(df[[1]])', kernel='R')
        assert 'integer' in notebook.check_output('class(df[[2]])', kernel='R')

    def test_put_dataframe(self, notebook):
        notebook.call('%put mtcars', kernel='R')
        assert '32' == notebook.check_output('mtcars.shape[0]', kernel='SoS')
        assert '11' == notebook.check_output('mtcars.shape[1]', kernel='SoS')
        assert "'Mazda RX4'" == notebook.check_output(
            'mtcars.index[0]', kernel='SoS')

    def test_get_dict_with_special_keys(self, notebook):
        output = self.get_from_SoS(
            notebook, "{'11111': 1, '_1111': 'a', 11112: 2, (1,2): 3}")
        assert '$X11111' in output and '$X_1111' in output and '$X11112' in output and '$X_1__2_' in output
